function [sol,flag,runt,perft,emess]=DPA(xo,yo,xf,yf,ndx,ndy,wall,displ)

% Diffusion pathfinding algorithm (DPA) - Version 1.0
% Date: 02/03/2016
% Author: Benjamin Ivorra
% MOMAT research group, IMI-Institute & Applied Mathematics Department. Complutense University of Madrid, Spain.
% Contact for support: ivorra@mat.ucm.es

% Help guide:

% DPA is a 2D pathfinding algorithm based on a diffusion technique starting from the target point up to the initial point of the path.
% The environment is coded by using a rectangular grid.

% Input variables:

% ndx   = 2D grid size in the X-axis - Natural number greater than 5
% ndy   = 2D grid size in the Y-axis - Natural number greater than 5
% xo    = initial position in the X-axis - Natural number in [2,ndx-1]
% yo    = initial position in the X-axis - Natural number in [2,ndx-1]
% xf    = final position in the X-axis - Natural number in [2,ndx-1]
% yf    = final position in the Y-axis - Natural number in [2,ndx-1]
% wall  = an integer (ndx,ndy) matrix corresponding to the wall that path cannot
% cross: 0= no wall | 1= wall
% displ = define if graphical and textual outputs should be displayed on the
% screen - 1 = On | 0 = Off

% Output variables:

% sol    = a matrix containing the solution path: 1st column= X coordinate | 2nd
% column = Y coordinate. If no solution was found this matrix is void.
% flag   = an integer returning: 0= No solution was found | 1= Solution found |2= Fatal error
% runt   = Running time of the DPA without output generation (in secs)
% perft  = Performance time of the DPA defined as: nodes in the wall matrix
% equal to 0/runt (in Nodes per secs).
% emess  = a text containing the fatal error message

try
    
    wall=wall';
    sol=[];
    flag=0;
    sf=0;
    emess='';
    runt=Inf;
    perft=Inf;
    
    if displ==1
        
        disp(' ')
        disp('***********************************************')
        disp('* Diffusion pathfinding algorithm (DPA) V 1.0 *')
        disp('***********************************************')
        disp(' ')
        disp(['Author: Benjamin Ivorra '])
        disp(['MOMAT research group, IMI-Institute & Applied Mathematics Department. Complutense University of Madrid, Spain.'])
        disp(['Contact for support: ivorra@mat.ucm.es'])
        disp(' ')
        disp('1- Drawing environment')
        
        % Drawing the problem
        figure(1)
        clf
        hold on
        surface(wall,'facecolor','texturemap','edgecolor','none')
        c=colormap('gray');
        colormap(1-c)
        plot(xo,yo,'rs','linewidth',3)
        plot(xf,yf,'gs','linewidth',3)
        xlabel('X')
        ylabel('Y')
        axis tight
        pause(.05)
    end
    
    if displ==1
        tic
        disp(' ')
        disp('2- Solving pathfinding problem')
    end
    
    temp=zeros(ndy,ndx);
    temp(xo,yo)=0;
    temp(yf,xf)=-1;
    
    % Diffusion model
    temp2=zeros(ndy,ndx);
    for tt=1:ndx*ndy
        temp2(2:ndy-1,2:ndx-1)=(temp(2:ndy-1,2:ndx-1)+temp((2:ndy-1)-1,2:ndx-1)+temp((2:ndy-1)+1,2:ndx-1) ...
            +temp(2:ndy-1,(2:ndx-1)-1)+temp(2:ndy-1,(2:ndx-1)+1))/5;
        temp2(wall==1)=0;
        temp2(yf,xf)=-1;
        temp=temp2/max(max(abs(temp2)));
        
        if temp(xo,yo)~=0
            break
        end
        
    end
    temp=temp';
    
    if tt==ndx*ndy
        runt=toc;
        if displ==1
            disp(' ')
            disp(['Solution not found'])
            disp(' ')
            disp(['Running time: ' num2str(runt) ' (s)'])
        end
    else
        
        % Building path
        px(1)=xo;
        py(1)=yo;
        pz(1)=temp(px(1),py(1));
        diry=[-1,0,+1;-1,0,+1;-1,0,+1];
        dirx=-1*[1,1,1;0,0,0;-1,-1,-1];
        for tt=1:ndx*ndy
            matm=temp(px(tt)-1:px(tt)+1,py(tt)-1:py(tt)+1);
            mm=min(min(matm));
            plusx=dirx(matm==mm);
            plusy=diry(matm==mm);
            plusx=plusx(1);plusy=plusy(1);
            px(tt+1)=min(max(px(tt)+plusx,2),ndx-1);
            py(tt+1)=min(max(py(tt)+plusy,2),ndy-1);
            if px(tt)==xf&py(tt)==yf
                sf=1;
                break
            end
            if (px(tt)-px(tt+1)==0)&(py(tt)-py(tt+1)==0)
                sf=0;
                break
            end
        end
        if sf==1
            runt=toc;
            perft=round(sum(sum(temp~=0))/runt);
            if displ==1
                disp(' ')
                disp(['Running time: ' num2str(runt) ' (s)'])
                disp(['Performance: ' num2str(perft) ' (Nodes/s) '])
                disp(' ')
                disp('3- Building outputs')
                figure(1)
                plot(px,py,'b','linewidth',2)
                plot(xo,yo,'rs','linewidth',3)
                plot(xf,yf,'gs','linewidth',3)
            end
            sol=[px',py'];
            flag=1;
        else
            if displ==1
                disp(' ')
                disp(['Solution not found'])
                disp(' ')
                tot=toc;
                disp(['Running time: ' num2str(tot) ' (s)'])
            end
        end
    end
    if displ==1
        
        disp(' ')
        disp('End of the algorithm')
        disp(' ')
    end
catch expm
    flag=2;
    emess=expm.message;
    if displ==1
        disp(' ')
        disp('Fatal Error:')
        disp(' ')
        disp(emess)
    end
    sol=[];
end
end



